// (c) by Stefan Roettger, licensed under LGPL 2.1

#include "minibase.h"
#include "minimath.h"

#include "minibrick.h"

#include "ministrip.h"

// default constructor
ministrip::ministrip(int colcomps,int nrmcomps,int texcomps)
   : miniARB()
   {init(colcomps,nrmcomps,texcomps);}

// copy constructor
ministrip::ministrip(const ministrip &strip)
   : miniARB()
   {
   init(strip.COLCOMPS,strip.NRMCOMPS,strip.TEXCOMPS,strip.MAXSIZE);

   SIZE=strip.SIZE;

   memcpy(ENDARRAY,strip.ENDARRAY,SIZE*sizeof(BOOLINT));
   memcpy(VTXARRAY,strip.VTXARRAY,3*SIZE*sizeof(float));

   if (COLARRAY!=NULL)
      memcpy(COLARRAY,strip.COLARRAY,COLCOMPS*SIZE*sizeof(float));

   if (NRMARRAY!=NULL)
      memcpy(NRMARRAY,strip.NRMARRAY,NRMCOMPS*SIZE*sizeof(float));

   if (TEXARRAY!=NULL)
      memcpy(TEXARRAY,strip.TEXARRAY,TEXCOMPS*SIZE*sizeof(float));

   BBOXMIN=strip.BBOXMIN;
   BBOXMAX=strip.BBOXMAX;

   COLAMIN=strip.COLAMIN;
   COLAMAX=strip.COLAMAX;
   }

// destructor
ministrip::~ministrip()
   {
   free(ENDARRAY);
   free(VTXARRAY);

   if (COLARRAY!=NULL) free(COLARRAY);
   if (NRMARRAY!=NULL) free(NRMARRAY);
   if (TEXARRAY!=NULL) free(TEXARRAY);
   }

// assignment operator
ministrip& ministrip::operator = (const ministrip &strip)
   {
   clear();

   init(strip.COLCOMPS,strip.NRMCOMPS,strip.TEXCOMPS,strip.MAXSIZE);

   SIZE=strip.SIZE;

   memcpy(ENDARRAY,strip.ENDARRAY,SIZE*sizeof(BOOLINT));
   memcpy(VTXARRAY,strip.VTXARRAY,3*SIZE*sizeof(float));

   if (COLARRAY!=NULL)
      memcpy(COLARRAY,strip.COLARRAY,COLCOMPS*SIZE*sizeof(float));

   if (NRMARRAY!=NULL)
      memcpy(NRMARRAY,strip.NRMARRAY,NRMCOMPS*SIZE*sizeof(float));

   if (TEXARRAY!=NULL)
      memcpy(TEXARRAY,strip.TEXARRAY,TEXCOMPS*SIZE*sizeof(float));

   BBOXMIN=strip.BBOXMIN;
   BBOXMAX=strip.BBOXMAX;

   COLAMIN=strip.COLAMIN;
   COLAMAX=strip.COLAMAX;

   return(*this);
   }

// init strip
void ministrip::init(int colcomps,int nrmcomps,int texcomps,unsigned int maxsize)
   {
   int i;

   if (maxsize==0) ERRORMSG();
   if (colcomps!=0 && colcomps!=3 && colcomps!=4) ERRORMSG();
   if (nrmcomps!=0 && nrmcomps!=3) ERRORMSG();
   if (texcomps<0 || texcomps>4) ERRORMSG();

   COLCOMPS=colcomps;
   NRMCOMPS=nrmcomps;
   TEXCOMPS=texcomps;

   MAXSIZE=maxsize;

   if ((ENDARRAY=(BOOLINT *)malloc(MAXSIZE*sizeof(BOOLINT)))==NULL) MEMERROR();
   if ((VTXARRAY=(float *)malloc(3*MAXSIZE*sizeof(float)))==NULL) MEMERROR();

   if (COLCOMPS==0) COLARRAY=NULL;
   else
      if ((COLARRAY=(float *)malloc(COLCOMPS*MAXSIZE*sizeof(float)))==NULL) MEMERROR();

   if (NRMCOMPS==0) NRMARRAY=NULL;
   else
      if ((NRMARRAY=(float *)malloc(NRMCOMPS*MAXSIZE*sizeof(float)))==NULL) MEMERROR();

   if (TEXCOMPS==0) TEXARRAY=NULL;
   else
      if ((TEXARRAY=(float *)malloc(TEXCOMPS*MAXSIZE*sizeof(float)))==NULL) MEMERROR();

   SIZE=0;

   BBOXMIN=miniv3d(MAXFLOAT,MAXFLOAT,MAXFLOAT);
   BBOXMAX=miniv3d(-MAXFLOAT,-MAXFLOAT,-MAXFLOAT);

   COLR=1.0f;
   COLG=1.0f;
   COLB=1.0f;
   COLA=1.0f;

   COLAMIN=1.0f;
   COLAMAX=0.0f;

   NRMX=0.0f;
   NRMY=1.0f;
   NRMZ=0.0f;

   TEXX=0.0f;
   TEXY=0.0f;
   TEXZ=0.0f;
   TEXW=1.0f;

   COPYVTX=0;

   SCALE=1.0f;

   for (i=0; i<16; i++) MTX[i]=0.0;
   for (i=0; i<4; i++) MTX[i+4*i]=1.0;
   MTXSET=FALSE;

   for (i=0; i<16; i++) TEXMTX[i]=0.0;
   for (i=0; i<4; i++) TEXMTX[i+4*i]=1.0;
   TEXMTXSET=FALSE;

   ZSCALE=1.0f;

   AUTODISABLECULLING=FALSE;

   USESHADER=-1;
   }

// reinit strip
void ministrip::reinit(int colcomps,int nrmcomps,int texcomps,unsigned int maxsize)
   {
   if (maxsize==0) maxsize=1;

   free(ENDARRAY);
   free(VTXARRAY);

   if (COLARRAY!=NULL) free(COLARRAY);
   if (NRMARRAY!=NULL) free(NRMARRAY);
   if (TEXARRAY!=NULL) free(TEXARRAY);

   init(colcomps,nrmcomps,texcomps,maxsize);
   }

// clear strip
void ministrip::clear()
   {
   SIZE=0;

   BBOXMIN=miniv3d(MAXFLOAT,MAXFLOAT,MAXFLOAT);
   BBOXMAX=miniv3d(-MAXFLOAT,-MAXFLOAT,-MAXFLOAT);

   COLAMIN=1.0f;
   COLAMAX=0.0f;

   COPYVTX=0;
   }

// begin next triangle strip
void ministrip::beginstrip()
   {
   if (SIZE>0)
      {
      addvtx();
      ENDARRAY[SIZE-1]=TRUE;
      COPYVTX=1;
      }
   }

// set color of next vertex
void ministrip::setcol(const float r,const float g,const float b,float a)
   {
   COLR=r;
   COLG=g;
   COLB=b;
   COLA=a;

   if (a<COLAMIN) COLAMIN=a;
   if (a>COLAMAX) COLAMAX=a;
   }

void ministrip::setcol(const miniv3f &rgb)
   {setcol(rgb.x,rgb.y,rgb.z,1.0);}

void ministrip::setcol(const miniv4f &rgba)
   {setcol(rgba.x,rgba.y,rgba.z,rgba.w);}

// set normal of next vertex
void ministrip::setnrm(const float nx,const float ny,const float nz)
   {
   NRMX=nx;
   NRMY=ny;
   NRMZ=nz;
   }

void ministrip::setnrm(const miniv3f &n)
   {setnrm(n.x,n.y,n.z);}

// set normal from three points
void ministrip::setnrm(const miniv3d &v1,const miniv3d &v2,const miniv3d &v3)
   {setnrm((v2-v1)/(v3-v1));}

// set tex coords of next vertex
void ministrip::settex(const float tx,const float ty,const float tz,float tw)
   {
   TEXX=tx;
   TEXY=ty;
   TEXZ=tz;
   TEXW=tw;
   }

void ministrip::settex(const miniv3f &t)
   {settex(t.x,t.y,t.z);}

void ministrip::settex(const miniv4f &t)
   {settex(t.x,t.y,t.z,t.w);}

// add one vertex
void ministrip::addvtx(const float x,const float y,const float z)
   {
   VTXX=x;
   VTXY=y;
   VTXZ=z;

   addvtx();

   if (COPYVTX!=0)
      {
      addvtx();
      COPYVTX=0;
      }
   }

void ministrip::addvtx(const miniv3f &v)
   {addvtx(v.x,v.y,v.z);}

// add one vertex
void ministrip::addvtx()
   {
   float *ptr;

   if (SIZE>=MAXSIZE)
      {
      if (MAXSIZE<2048) MAXSIZE*=2;
      else MAXSIZE=MAXSIZE+(MAXSIZE>>1);

      if ((ENDARRAY=(BOOLINT *)realloc(ENDARRAY,MAXSIZE*sizeof(BOOLINT)))==NULL) MEMERROR();
      if ((VTXARRAY=(float *)realloc(VTXARRAY,3*MAXSIZE*sizeof(float)))==NULL) MEMERROR();

      if (COLARRAY!=NULL)
         if ((COLARRAY=(float *)realloc(COLARRAY,COLCOMPS*MAXSIZE*sizeof(float)))==NULL) MEMERROR();

      if (NRMARRAY!=NULL)
         if ((NRMARRAY=(float *)realloc(NRMARRAY,NRMCOMPS*MAXSIZE*sizeof(float)))==NULL) MEMERROR();

      if (TEXARRAY!=NULL)
         if ((TEXARRAY=(float *)realloc(TEXARRAY,TEXCOMPS*MAXSIZE*sizeof(float)))==NULL) MEMERROR();
      }

   if (VTXX<BBOXMIN.x) BBOXMIN.x=VTXX;
   if (VTXX>BBOXMAX.x) BBOXMAX.x=VTXX;

   if (VTXY<BBOXMIN.y) BBOXMIN.y=VTXY;
   if (VTXY>BBOXMAX.y) BBOXMAX.y=VTXY;

   if (VTXZ<BBOXMIN.z) BBOXMIN.z=VTXZ;
   if (VTXZ>BBOXMAX.z) BBOXMAX.z=VTXZ;

   ENDARRAY[SIZE]=FALSE;

   ptr=&VTXARRAY[3*SIZE];

   *ptr++=VTXX;
   *ptr++=VTXY;
   *ptr++=VTXZ;

   if (COLARRAY!=NULL)
      {
      ptr=&COLARRAY[COLCOMPS*SIZE];

      *ptr++=COLR;
      *ptr++=COLG;
      *ptr++=COLB;

      if (COLCOMPS==4) *ptr=COLA;
      }

   if (NRMARRAY!=NULL)
      {
      ptr=&NRMARRAY[NRMCOMPS*SIZE];

      *ptr++=NRMX;
      *ptr++=NRMY;
      *ptr++=NRMZ;
      }

   if (TEXARRAY!=NULL)
      {
      ptr=&TEXARRAY[TEXCOMPS*SIZE];

      *ptr++=TEXX;

      if (TEXCOMPS>1) *ptr++=TEXY;
      if (TEXCOMPS>2) *ptr++=TEXZ;
      if (TEXCOMPS>3) *ptr=TEXW;
      }

   SIZE++;
   }

// set scale
void ministrip::setscale(float scale)
   {SCALE=scale;}

// set matrix
void ministrip::setmatrix(double mtx[16])
   {
   int i;

   for (i=0; i<16; i++) MTX[i]=mtx[i];
   MTXSET=TRUE;
   }

// set texture matrix
void ministrip::settexmatrix(double texmtx[16])
   {
   int i;

   for (i=0; i<16; i++) TEXMTX[i]=texmtx[i];
   TEXMTXSET=TRUE;
   }

// set Z-scaling
void ministrip::setZscale(float zscale)
   {ZSCALE=zscale;}

// auto disable culling
void ministrip::autodisableculling(BOOLINT on)
   {AUTODISABLECULLING=on;}

// set actual shader
void ministrip::useshader(int num)
   {USESHADER=num;}

// get actual shader
int ministrip::getshader()
   {return(USESHADER);}

// check whether or not the geometry is fully opaque
BOOLINT ministrip::isopaque()
   {
   if (COLCOMPS<4) return(TRUE);

   return(COLAMIN>=1.0f);
   }

// render triangle strips
void ministrip::render(int wocolor,int wonrm,int wotex)
   {
   int shader;

   BOOLINT culling;

   shader=USESHADER;
   if (shader<0) shader=getglobalshader();

   if (wocolor==0)
      if (COLARRAY==NULL) color(COLR,COLG,COLB,COLA);

   if (wonrm==0)
      if (NRMARRAY==NULL) normal(NRMX,NRMY,NRMZ);

   if (MTXSET || SCALE!=1.0f)
      {
      mtxpush();
      mtxmult(MTX);
      mtxscale(SCALE,SCALE,SCALE);
      }

   if (ZSCALE!=1.0f)
      {
      mtxproj();
      mtxpush();
      mtxscale(ZSCALE,ZSCALE,ZSCALE);
      mtxmodel();
      }

   if (shader>=0)
      {
      enablevtxshader(shader);
      enablepixshader(shader);
      }

   vertexarray(VTXARRAY);
   if (wocolor==0) colorarray(COLARRAY,COLCOMPS);
   if (wonrm==0) normalarray(NRMARRAY);

   if (wotex==0)
      if (TEXARRAY!=NULL)
         {
         if (TEXMTXSET)
            {
            mtxtex();
            mtxpush();
            mtxmult(TEXMTX);
            mtxmodel();
            }

         texclientunit(0);
         texcoordarray(TEXARRAY,TEXCOMPS);
         }
      else texcoordarray(NULL);

   if (AUTODISABLECULLING) culling=getculling();
   else culling=FALSE;

   if (culling) disableculling();
   rendergeo(wocolor,wonrm,wotex);
   if (culling) enableOLDculling();

   vertexarray(NULL);
   if (wocolor==0) colorarray(NULL);
   if (wonrm==0) normalarray(NULL);

   if (wotex==0)
      if (TEXARRAY!=NULL)
         {
         texcoordarray(NULL);

         if (TEXMTXSET)
            {
            mtxtex();
            mtxpop();
            mtxmodel();
            }
         }

   if (shader>=0)
      {
      disablevtxshader(shader);
      disablepixshader(shader);
      }

   if (ZSCALE!=1.0f)
      {
      mtxproj();
      mtxpop();
      mtxmodel();
      }

   if (MTXSET || SCALE!=1.0f) mtxpop();
   }

// render triangle strips with multi-pass method for unordered semi-transparent geometry
void ministrip::rendermultipass(int pass,
                                int passes,
                                float ambient,
                                float bordercontrol,float centercontrol,float colorcontrol,
                                float bordercontrol2,float centercontrol2,float colorcontrol2,
                                float stripewidth,float stripeoffset,
                                float stripedx,float stripedy,float stripedz,
                                int correctz)
   {
   static minisurf surf;

   int dorender;

   if (passes<1 || passes>4) ERRORMSG();

   // enable external multi-pass shader
   dorender=surf.setextstate(1,
                             pass,passes,
                             ambient,
                             bordercontrol,centercontrol,colorcontrol,
                             bordercontrol2,centercontrol2,colorcontrol2,
                             stripewidth,stripeoffset,stripedx,stripedy,stripedz,
                             correctz);

   // render strips with external multi-pass shader
   if (dorender!=0) render();

   // disable external multi-pass shader
   surf.setextstate(0,
                    pass,passes,
                    ambient,
                    bordercontrol,centercontrol,colorcontrol,
                    bordercontrol2,centercontrol2,colorcontrol2,
                    stripewidth,stripeoffset,stripedx,stripedy,stripedz,
                    correctz);
   }

// render triangle strip with multi-pass method
void ministrip::rendermulti(int passes,
                            float ambient,
                            float bordercontrol,float centercontrol,float colorcontrol,
                            float bordercontrol2,float centercontrol2,float colorcontrol2,
                            float stripewidth,float stripeoffset,
                            float stripedx,float stripedy,float stripedz,
                            int correctz)
   {
   int i;

   int globalshader,shader;

   // disable regular shader
   globalshader=getglobalshader();
   useglobalshader(-1);
   shader=getshader();
   useshader(-1);

   // multi-pass rendering
   for (i=minisurf::FIRST_RENDER_PHASE; i<=minisurf::LAST_RENDER_PHASE; i++)
      rendermultipass(i,passes,
                      ambient,
                      bordercontrol,centercontrol,colorcontrol,
                      bordercontrol2,centercontrol2,colorcontrol2,
                      stripewidth,stripeoffset,
                      stripedx,stripedy,stripedz,
                      correctz);

   // enable previous shader
   useglobalshader(globalshader);
   useshader(shader);
   }

// render geometry
void ministrip::rendergeo(int /*wocolor*/,int /*wonrm*/,int /*wotex*/)
   {renderstrip(0,SIZE);}

// render bbox
void ministrip::renderbbox()
   {
   renderline(miniv3d(BBOXMIN.x,BBOXMIN.y,BBOXMIN.z),miniv3d(BBOXMAX.x,BBOXMIN.y,BBOXMIN.z));
   renderline(miniv3d(BBOXMIN.x,BBOXMAX.y,BBOXMIN.z),miniv3d(BBOXMAX.x,BBOXMAX.y,BBOXMIN.z));
   renderline(miniv3d(BBOXMIN.x,BBOXMIN.y,BBOXMAX.z),miniv3d(BBOXMAX.x,BBOXMIN.y,BBOXMAX.z));
   renderline(miniv3d(BBOXMIN.x,BBOXMAX.y,BBOXMAX.z),miniv3d(BBOXMAX.x,BBOXMAX.y,BBOXMAX.z));
   renderline(miniv3d(BBOXMIN.x,BBOXMIN.y,BBOXMIN.z),miniv3d(BBOXMIN.x,BBOXMAX.y,BBOXMIN.z));
   renderline(miniv3d(BBOXMAX.x,BBOXMIN.y,BBOXMIN.z),miniv3d(BBOXMAX.x,BBOXMAX.y,BBOXMIN.z));
   renderline(miniv3d(BBOXMIN.x,BBOXMIN.y,BBOXMAX.z),miniv3d(BBOXMIN.x,BBOXMAX.y,BBOXMAX.z));
   renderline(miniv3d(BBOXMAX.x,BBOXMIN.y,BBOXMAX.z),miniv3d(BBOXMAX.x,BBOXMAX.y,BBOXMAX.z));
   renderline(miniv3d(BBOXMIN.x,BBOXMIN.y,BBOXMIN.z),miniv3d(BBOXMIN.x,BBOXMIN.y,BBOXMAX.z));
   renderline(miniv3d(BBOXMAX.x,BBOXMIN.y,BBOXMIN.z),miniv3d(BBOXMAX.x,BBOXMIN.y,BBOXMAX.z));
   renderline(miniv3d(BBOXMIN.x,BBOXMAX.y,BBOXMIN.z),miniv3d(BBOXMIN.x,BBOXMAX.y,BBOXMAX.z));
   renderline(miniv3d(BBOXMAX.x,BBOXMAX.y,BBOXMIN.z),miniv3d(BBOXMAX.x,BBOXMAX.y,BBOXMAX.z));
   }

// get bounding box
void ministrip::getbbox(miniv3d &bboxmin,miniv3d &bboxmax) const
   {
   bboxmin=BBOXMIN;
   bboxmax=BBOXMAX;
   }

// get bounding sphere
void ministrip::getbsphere(miniv3d &center,double &radius2) const
   {
   miniv3d bboxmin,bboxmax;

   getbbox(bboxmin,bboxmax);

   center=0.5*(bboxmin+bboxmax);
   radius2=(0.5*(bboxmax-bboxmin)).getlength2();
   }

// shoot a ray and return the distance to the closest triangle
double ministrip::shoot(const miniv3d &o,const miniv3d &d,double mindist) const
   {
   unsigned int i;

   float *ptr;
   miniv3f v1,v2,v3;

   double dist,result;

   result=MAXFLOAT;

   if (!MTXSET)
      if (itest_ray_bbox(o,d,0.5*(BBOXMIN+BBOXMAX),0.5*(BBOXMAX-BBOXMIN)))
         {
         ptr=VTXARRAY;

         for (i=0; i<SIZE; i++)
            {
            v1.x=*ptr++;
            v1.y=*ptr++;
            v1.z=*ptr++;

            if (i>=2)
               {
               dist=ray_triangle_dist(o,d,v1,v2,v3);
               if (dist!=MAXFLOAT)
                  if (dist>=mindist)
                     if (dist<result) result=dist;

               v3=v2;
               v2=v1;
               }
            else if (i>=1)
               {
               v3=v2;
               v2=v1;
               }
            else v2=v1;
            }
         }

   return(result);
   }

// shoot a ray and return the distance to the bounding sphere
double ministrip::shootbsphere(const miniv3d &o,const miniv3d &d,double mindist) const
   {
   miniv3d b;
   double r,r2;
   double l;

   getbsphere(b,r2);
   r=sqrt(r2);

   l=intersect_ray_ellipsoid(o,d,b,r,r,r);

   return((l<mindist)?-MAXFLOAT:l);
   }

// serialization
ministring ministrip::to_string() const
   {
   ministring info("ministrip");

   info.append("(");
   info.append_uint(SIZE);
   info.append(",");
   info.append_int(COLCOMPS);
   info.append(",");
   info.append_int(NRMCOMPS);
   info.append(",");
   info.append_int(TEXCOMPS);
   info.append(")");

   info.append("end[");
   info.append_array(ENDARRAY,SIZE);
   info.append("]");

   info.append("vertex[");
   info.append_array(VTXARRAY,3*SIZE);
   info.append("]");

   if (COLCOMPS>0)
      {
      info.append("color[");
      info.append_array(COLARRAY,COLCOMPS*SIZE);
      info.append("]");
      }

   if (NRMCOMPS>0)
      {
      info.append("normal[");
      info.append_array(NRMARRAY,NRMCOMPS*SIZE);
      info.append("]");
      }

   if (TEXCOMPS>0)
      {
      info.append("texcoord[");
      info.append_array(TEXARRAY,TEXCOMPS*SIZE);
      info.append("]");
      }

   info.append("(");
   info.append(BBOXMIN.to_string());
   info.append(",");
   info.append(BBOXMAX.to_string());
   info.append(",");
   info.append_float(COLAMIN);
   info.append(",");
   info.append_float(COLAMAX);
   info.append(")");

   return(info);
   }

// deserialization
void ministrip::from_string(ministring &info)
   {
   unsigned int size;
   int colcomps,nrmcomps,texcomps;

   if (info.startswith("ministrip"))
      {
      info=info.tail("ministrip(");
      size=info.prefix(",").value_uint();
      info=info.tail(",");
      colcomps=info.prefix(",").value_int();
      info=info.tail(",");
      nrmcomps=info.prefix(",").value_int();
      info=info.tail(",");
      texcomps=info.prefix(")").value_int();
      info=info.tail(")");

      reinit(colcomps,nrmcomps,texcomps,size);

      info=info.tail("end[");
      info.extract_array(ENDARRAY,size);
      info.startat(size);
      info=info.tail("]");

      info=info.tail("vertex[");
      info.extract_array(VTXARRAY,3*size);
      info.startat(3*size);
      info=info.tail("]");

      if (colcomps>0)
         {
         info=info.tail("color[");
         info.extract_array(VTXARRAY,colcomps*size);
         info.startat(colcomps*size);
         info=info.tail("]");
         }

      if (nrmcomps>0)
         {
         info=info.tail("normal[");
         info.extract_array(NRMARRAY,nrmcomps*size);
         info.startat(nrmcomps*size);
         info=info.tail("]");
         }

      if (texcomps>0)
         {
         info=info.tail("texcoord[");
         info.extract_array(TEXARRAY,texcomps*size);
         info.startat(texcomps*size);
         info=info.tail("]");
         }

      SIZE=size;

      info=info.tail("(");
      BBOXMIN.from_string(info);
      info=info.tail(",");
      BBOXMAX.from_string(info);
      info=info.tail(",");
      COLAMIN=info.prefix(",").value();
      info=info.tail(",");
      COLAMAX=info.prefix(")").value();
      info=info.tail(")");
      }
   }

// save to file
void ministrip::save(const char *filename)
   {writeministring(filename,to_string());}

// load from file
BOOLINT ministrip::load(const char *filename)
   {
   ministring info;

   info=readministring(filename);

   if (!info.empty())
      {
      clear();
      from_string(info);

      return(TRUE);
      }

   return(FALSE);
   }

// get index range of next consecutive strip
BOOLINT ministrip::getnextrange(unsigned int &start,unsigned int &stop)
   {
   if (stop>0) start=stop=stop+3;

   while (stop+1<SIZE)
      if (ENDARRAY[stop+1]) break;
      else stop++;

   return(stop<SIZE);
   }

// write strip to GEO file format
void ministrip::writeGEOfile(const char *filename)
   {
   unsigned int i;

   unsigned int start,stop;

   FILE *outfile;

   // open output file
   if ((outfile=fopen(filename,"wb"))==NULL) return;

   // write GEO header
   fprintf(outfile,"geo 1.0\n"); // GEO format: magic identifier
   fprintf(outfile,"%d %d %d\n",COLCOMPS,NRMCOMPS,TEXCOMPS); // GEO format: number of attributes

   // write GEO body
   start=stop=0;
   while (getnextrange(start,stop))
      {
      fprintf(outfile,"%d\n",stop-start+1);

      for (i=start; i<=stop; i++)
         {
         fprintf(outfile,"%g %g %g",
                 VTXARRAY[3*i],VTXARRAY[3*i+1],VTXARRAY[3*i+2]);

         if (COLCOMPS==3 || COLCOMPS==4)
            {
            fprintf(outfile," %g %g %g",
                    COLARRAY[COLCOMPS*i],COLARRAY[COLCOMPS*i+1],COLARRAY[COLCOMPS*i+2]);

            if (COLCOMPS==4)
               fprintf(outfile," %g",
                       COLARRAY[COLCOMPS*i+3]);
            }

         if (NRMCOMPS==3)
            fprintf(outfile," %g %g %g",
                    NRMARRAY[3*i],NRMARRAY[3*i+1],NRMARRAY[3*i+2]);

         if (TEXCOMPS>0)
            {
            fprintf(outfile," %g",
                    TEXARRAY[TEXCOMPS*i]);

            if (TEXCOMPS>1)
               fprintf(outfile," %g",
                    TEXARRAY[TEXCOMPS*i+1]);

            if (TEXCOMPS>2)
               fprintf(outfile," %g",
                       TEXARRAY[TEXCOMPS*i+2]);

            if (TEXCOMPS>3)
               fprintf(outfile," %g",
                       TEXARRAY[TEXCOMPS*i+3]);
            }

         fprintf(outfile,"\n");
         }
      }

   // close output file
   fclose(outfile);
   }

// read strip from GEO file format
BOOLINT ministrip::readGEOfile(const char *filename)
   {
   unsigned int i;

   FILE *file;

   float version;

   unsigned int count;

   float x,y,z;
   float red,green,blue,alpha;
   float nx,ny,nz;
   float s,t,r,w;

   // open input file
   if ((file=fopen(filename,"rb"))==NULL) return(FALSE);

   // read GEO format: magic identifier
   if (fscanf(file,"geo %g\n",&version)!=1)
      {
      fclose(file);
      return(FALSE);
      }
   else if (version!=1.0f)
      {
      fclose(file);
      return(FALSE);
      }

   // read GEO format: number of attributes
   if (fscanf(file,"%d %d %d\n",&COLCOMPS,&NRMCOMPS,&TEXCOMPS)!=3)
      {
      fclose(file);
      return(FALSE);
      }
   else reinit(COLCOMPS,NRMCOMPS,TEXCOMPS);

   // read GEO body
   while (fscanf(file,"%d",&count)==1)
      {
      if (count==0) break;

      beginstrip();

      for (i=0; i<count; i++)
         {
         if (fscanf(file,"%g %g %g",&x,&y,&z)!=3) break;

         if (COLCOMPS==3 || COLCOMPS==4)
            {
            red=green=blue=alpha=0.0f;

            if (fscanf(file,"%g %g %g",&red,&green,&blue)!=3) break;

            if (COLCOMPS==4)
               if (fscanf(file,"%g",&alpha)!=1) break;

            setcol(red,green,blue,alpha);
            }

         if (NRMCOMPS==3)
            {
            if (fscanf(file,"%g %g %g",&nx,&ny,&nz)!=3) break;

            setnrm(nx,ny,nz);
            }

         if (TEXCOMPS>0)
            {
            s=t=r=w=0.0f;

            if (fscanf(file,"%g",&s)!=1) break;

            if (TEXCOMPS>1)
               if (fscanf(file,"%g",&t)!=1) break;

            if (TEXCOMPS>2)
               if (fscanf(file,"%g",&r)!=1) break;

            if (TEXCOMPS>3)
               if (fscanf(file,"%g",&w)!=1) break;

            settex(s,t,r,w);
            }

         addvtx(x,y,z);
         }
      }

   // close input file
   fclose(file);

   return(TRUE);
   }

// write strip to PLY file format
void ministrip::writePLYfile(const char *filename)
   {
   unsigned int i;

   unsigned int vertices,strips;
   unsigned int start,stop;
   unsigned int count;

   FILE *outfile;

   // count strips
   vertices=strips=0;
   start=stop=0;
   while (getnextrange(start,stop))
      {
      count=stop-start+1;

      if (count==3 || count==4)
         {
         vertices+=count;
         strips++;
         }
      }

   // open output file
   if ((outfile=fopen(filename,"wb"))==NULL) return;

   // write PLY header
   fprintf(outfile,"ply\n"); // ply format: magic identifier
   fprintf(outfile,"format ascii 1.0\n"); // ply format: format identifier
   fprintf(outfile,"comment created by libmini\n"); // ply format: libmini comment
   fprintf(outfile,"element vertex %u\n",vertices); // ply format: number of vertices
   fprintf(outfile,"property float x\n"); // ply format: property of vertex
   fprintf(outfile,"property float y\n"); // ply format: property of vertex
   fprintf(outfile,"property float z\n"); // ply format: property of vertex
   if (COLCOMPS==3 || COLCOMPS==4)
      {
      fprintf(outfile,"property float red\n"); // ply format: property of vertex
      fprintf(outfile,"property float green\n"); // ply format: property of vertex
      fprintf(outfile,"property float blue\n"); // ply format: property of vertex

      if (COLCOMPS==4)
         fprintf(outfile,"property float alpha\n"); // ply format: property of vertex
      }
   if (NRMCOMPS==3)
      {
      fprintf(outfile,"property float nx\n"); // ply format: property of vertex
      fprintf(outfile,"property float ny\n"); // ply format: property of vertex
      fprintf(outfile,"property float nz\n"); // ply format: property of vertex
      }
   fprintf(outfile,"element face %u\n",strips); // ply format: number of faces
   fprintf(outfile,"property list uchar int vertex_index\n"); // ply format: property of face
   fprintf(outfile,"end_header\n"); // ply format: end identifier

   // write PLY body (vertices)
   start=stop=0;
   while (getnextrange(start,stop))
      {
      count=stop-start+1;

      if (count==3 || count==4)
         for (i=start; i<=stop; i++)
            {
            fprintf(outfile,"%g %g %g",
                    VTXARRAY[3*i],VTXARRAY[3*i+1],VTXARRAY[3*i+2]);

            if (COLCOMPS==3 || COLCOMPS==4)
               {
               fprintf(outfile," %g %g %g",
                       COLARRAY[COLCOMPS*i],COLARRAY[COLCOMPS*i+1],COLARRAY[COLCOMPS*i+2]);

               if (COLCOMPS==4)
                  fprintf(outfile," %g",
                          COLARRAY[COLCOMPS*i+3]);
               }

            if (NRMCOMPS==3)
               fprintf(outfile," %g %g %g",
                       NRMARRAY[3*i],NRMARRAY[3*i+1],NRMARRAY[3*i+2]);

            fprintf(outfile,"\n");
            }
      }

   // write PLY body (faces)
   start=stop=0;
   vertices=0;
   while (getnextrange(start,stop))
      {
      count=stop-start+1;

      if (count==3)
         {
         fprintf(outfile,"%u",count);

         fprintf(outfile," %u",vertices);
         fprintf(outfile," %u",vertices+1);
         fprintf(outfile," %u",vertices+2);

         fprintf(outfile,"\n");

         vertices+=3;
         }
      else if (count==4)
         {
         fprintf(outfile,"%u",count);

         fprintf(outfile," %u",vertices);
         fprintf(outfile," %u",vertices+1);
         fprintf(outfile," %u",vertices+3);
         fprintf(outfile," %u",vertices+2);

         fprintf(outfile,"\n");

         vertices+=4;
         }
      }

   // close output file
   fclose(outfile);
   }

// convert strip from GEO to PLY file format
BOOLINT ministrip::convertGEO2PLYfile(const char *filename,const char *outname)
   {
   unsigned int i;

   FILE *file;
   long long tellpos;

   float version;
   int colcomps,nrmcomps,texcomps;

   unsigned int count;

   float x,y,z;
   float red,green,blue,alpha;
   float nx,ny,nz;
   float s,t,r,w;

   unsigned int strips,vertices;

   FILE *outfile;

   // open input file
   if ((file=fopen(filename,"rb"))==NULL) return(FALSE);

   // read GEO format: magic identifier
   if (fscanf(file,"geo %g\n",&version)!=1)
      {
      fclose(file);
      return(FALSE);
      }
   else if (version!=1.0f)
      {
      fclose(file);
      return(FALSE);
      }

   // read GEO format: number of attributes
   if (fscanf(file,"%d %d %d\n",&colcomps,&nrmcomps,&texcomps)!=3)
      {
      fclose(file);
      return(FALSE);
      }

   // remember seek position
   tellpos=ftell(file);

   strips=0;
   vertices=0;

   // scan GEO body
   while (fscanf(file,"%d",&count)==1)
      {
      if (count==0) break;

      if (count==3 || count==4)
         {
         strips++;
         vertices+=count;
         }

      for (i=0; i<count; i++)
         {
         if (fscanf(file,"%g %g %g",&x,&y,&z)!=3) break;

         if (colcomps==3 || colcomps==4)
            {
            red=green=blue=alpha=0.0f;

            if (fscanf(file,"%g %g %g",&red,&green,&blue)!=3) break;

            if (colcomps==4)
               if (fscanf(file,"%g",&alpha)!=1) break;
            }

         if (nrmcomps==3)
            if (fscanf(file,"%g %g %g",&nx,&ny,&nz)!=3) break;

         if (texcomps>0)
            {
            s=t=r=w=0.0f;

            if (fscanf(file,"%g",&s)!=1) break;

            if (texcomps>1)
               if (fscanf(file,"%g",&t)!=1) break;

            if (texcomps>2)
               if (fscanf(file,"%g",&r)!=1) break;

            if (texcomps>3)
               if (fscanf(file,"%g",&w)!=1) break;
            }
         }
      }

   // seek back to start
   if (fseek(file,tellpos,SEEK_SET)==-1)
      {
      fclose(file);
      return(FALSE);
      }

   // open output file
   if ((outfile=fopen(outname,"wb"))==NULL)
      {
      fclose(file);
      return(FALSE);
      }

   // write PLY header
   fprintf(outfile,"ply\n"); // ply format: magic identifier
   fprintf(outfile,"format ascii 1.0\n"); // ply format: format identifier
   fprintf(outfile,"comment created by libmini\n"); // ply format: libmini comment
   fprintf(outfile,"element vertex %u\n",vertices); // ply format: number of vertices
   fprintf(outfile,"property float x\n"); // ply format: property of vertex
   fprintf(outfile,"property float y\n"); // ply format: property of vertex
   fprintf(outfile,"property float z\n"); // ply format: property of vertex
   if (colcomps==3 || colcomps==4)
      {
      fprintf(outfile,"property float red\n"); // ply format: property of vertex
      fprintf(outfile,"property float green\n"); // ply format: property of vertex
      fprintf(outfile,"property float blue\n"); // ply format: property of vertex

      if (colcomps==4)
         fprintf(outfile,"property float alpha\n"); // ply format: property of vertex
      }
   if (nrmcomps==3)
      {
      fprintf(outfile,"property float nx\n"); // ply format: property of vertex
      fprintf(outfile,"property float ny\n"); // ply format: property of vertex
      fprintf(outfile,"property float nz\n"); // ply format: property of vertex
      }
   fprintf(outfile,"element face %u\n",strips); // ply format: number of faces
   fprintf(outfile,"property list uchar int vertex_index\n"); // ply format: property of face
   fprintf(outfile,"end_header\n"); // ply format: end identifier

   // scan GEO body and write PLY body (vertices)
   while (fscanf(file,"%d",&count)==1)
      {
      if (count==0) break;

      for (i=0; i<count; i++)
         {
         if (fscanf(file,"%g %g %g",&x,&y,&z)!=3) break;
         if (count==3 || count==4) fprintf(outfile,"%g %g %g",x,y,z);

         if (colcomps==3 || colcomps==4)
            {
            if (fscanf(file,"%g %g %g",&red,&green,&blue)!=3) break;
            if (count==3 || count==4) fprintf(outfile," %g %g %g",red,green,blue);

            if (colcomps==4)
               {
               if (fscanf(file,"%g",&alpha)!=1) break;
               if (count==3 || count==4) fprintf(outfile," %g",alpha);
               }
            }

         if (nrmcomps==3)
            {
            if (fscanf(file,"%g %g %g",&nx,&ny,&nz)!=3) break;
            if (count==3 || count==4) fprintf(outfile," %g %g %g",nx,ny,nz);
            }

         if (texcomps>0)
            {
            if (fscanf(file,"%g",&s)!=1) break;

            if (texcomps>1)
               if (fscanf(file,"%g",&t)!=1) break;

            if (texcomps>2)
               if (fscanf(file,"%g",&r)!=1) break;

            if (texcomps>3)
               if (fscanf(file,"%g",&w)!=1) break;
            }

         if (count==3 || count==4) fprintf(outfile,"\n");
         }
      }

   // seek back to start
   if (fseek(file,tellpos,SEEK_SET)==-1)
      {
      fclose(file);
      fclose(outfile);
      return(FALSE);
      }

   vertices=0;

   // scan GEO body and write PLY body (faces)
   while (fscanf(file,"%d",&count)==1)
      {
      if (count==0) break;

      if (count==3)
         {
         fprintf(outfile,"%d %u %u %u\n",3,vertices,vertices+1,vertices+2);
         vertices+=3;
         }
      else if (count==4)
         {
         fprintf(outfile,"%d %u %u %u %u\n",4,vertices,vertices+1,vertices+3,vertices+2);
         vertices+=4;
         }

      for (i=0; i<count; i++)
         {
         if (fscanf(file,"%g %g %g",&x,&y,&z)!=3) break;

         if (colcomps==3 || colcomps==4)
            {
            if (fscanf(file,"%g %g %g",&red,&green,&blue)!=3) break;

            if (colcomps==4)
               if (fscanf(file,"%g",&alpha)!=1) break;
            }

         if (nrmcomps==3)
            if (fscanf(file,"%g %g %g",&nx,&ny,&nz)!=3) break;

         if (texcomps>0)
            {
            if (fscanf(file,"%g",&s)!=1) break;

            if (texcomps>1)
               if (fscanf(file,"%g",&t)!=1) break;

            if (texcomps>2)
               if (fscanf(file,"%g",&r)!=1) break;

            if (texcomps>3)
               if (fscanf(file,"%g",&w)!=1) break;
            }
         }
      }

   // close output file
   fclose(outfile);

   // close input file
   fclose(file);

   return(TRUE);
   }
