// (c) by Stefan Roettger, licensed under LGPL 2.1

#ifndef MINISTRIP_H
#define MINISTRIP_H

#include "miniOGL.h"
#include "miniARB.h"

#include "miniv3f.h"
#include "miniv4f.h"
#include "miniv3d.h"
#include "miniv4d.h"

//! container class for triangle strips
class ministrip: public miniARB
   {
   public:

   //! default constructor
   ministrip(int colcomps=0,int nrmcomps=0,int texcomps=0);

   //! copy constructor
   ministrip(const ministrip &strip);

   //! destructor
   virtual ~ministrip();

   //! assignment operator
   ministrip& operator = (const ministrip &strip);

   //! clear strip
   void clear();

   //! check strip
   BOOLINT empty() {return(SIZE==0);}

   //! begin next triangle strip
   void beginstrip();

   //! set color of next vertex
   void setcol(const float r,const float g,const float b,float a=1.0f);
   void setcol(const miniv3f &rgb);
   void setcol(const miniv4f &rgba);

   //! set normal of next vertex
   void setnrm(const float nx,const float ny,const float nz);
   void setnrm(const miniv3f &n);

   //! set normal from three points
   void setnrm(const miniv3d &v1,const miniv3d &v2,const miniv3d &v3);

   //! set tex coords of next vertex
   void settex(const float tx,const float ty=0.0f,const float tz=0.0f,float tw=1.0f);
   void settex(const miniv3f &t);
   void settex(const miniv4f &t);

   //! add one vertex
   void addvtx(const float x,const float y,const float z);
   void addvtx(const miniv3f &v);

   //! check color array
   BOOLINT hascolor() const {return(COLCOMPS>0);}

   //! check normal array
   BOOLINT hasnormal() const {return(NRMCOMPS>0);}

   //! check tex coord array
   BOOLINT hastex() const {return(TEXCOMPS>0);}

   //! set scale
   void setscale(float scale);

   //! set matrix
   void setmatrix(double mtx[16]);

   //! set texture matrix
   void settexmatrix(double texmtx[16]);

   //! set Z-scaling
   void setZscale(float zscale);

   //! auto disable culling
   void autodisableculling(BOOLINT on=TRUE);

   //! set actual shader
   void useshader(int num=0);

   //! get actual shader
   int getshader();

   //! check whether or not the geometry is fully opaque
   BOOLINT isopaque();

   //! render triangle strips
   void render(int wocolor=0,int wonrm=0,int wotex=0);

   //! render triangle strips with multi-pass method for unordered semi-transparent geometry
   void rendermultipass(int pass,
                        int passes=4,
                        float ambient=0.1f,
                        float bordercontrol=1.0f,float centercontrol=1.0f,float colorcontrol=1.0f,
                        float bordercontrol2=1.0f,float centercontrol2=1.0f,float colorcontrol2=1.0f,
                        float stripewidth=1.0f,float stripeoffset=0.0f,
                        float stripedx=0.0f,float stripedy=0.0f,float stripedz=1.0f,
                        int correctz=0);

   //! render triangle strip with multi-pass method
   void rendermulti(int passes=4,
                    float ambient=0.1f,
                    float bordercontrol=1.0f,float centercontrol=1.0f,float colorcontrol=1.0f,
                    float bordercontrol2=1.0f,float centercontrol2=1.0f,float colorcontrol2=1.0f,
                    float stripewidth=1.0f,float stripeoffset=0.0f,
                    float stripedx=0.0f,float stripedy=0.0f,float stripedz=1.0f,
                    int correctz=0);

   //! render geometry
   virtual void rendergeo(int wocolor=0,int wonrm=0,int wotex=0);

   //! render bbox
   virtual void renderbbox();

   //! get array size (number of stored vertices)
   unsigned int getsize() {return(SIZE);}

   //! get vertex array
   float *getvtx() {return(VTXARRAY);}

   //! get reference to vertex array
   float **getvtxref() {return(&VTXARRAY);}

   //! get bounding box
   virtual void getbbox(miniv3d &bboxmin,miniv3d &bboxmax) const;

   //! get bounding sphere
   virtual void getbsphere(miniv3d &center,double &radius2) const;

   //! shoot a ray and return the distance to the closest triangle
   //! only the triangles with a positive distance are considered
   //! o is the origin of the ray, d is the ray direction
   //! a valid hit distance to be returned has to be no less than mindist
   //! a return value of MAXFLOAT indicates that there was no hit
   virtual double shoot(const miniv3d &o,const miniv3d &d,double mindist=0.0) const;

   //! shoot a ray and return the distance to the bounding sphere
   virtual double shootbsphere(const miniv3d &o,const miniv3d &d,double mindist=0.0) const;

   //! serialization
   ministring to_string() const;

   //! deserialization
   void from_string(ministring &info);

   //! save to file
   void save(const char *filename);

   //! load from file
   BOOLINT load(const char *filename);

   //! write strip to GEO file format
   void writeGEOfile(const char *filename);

   //! read strip from GEO file format
   BOOLINT readGEOfile(const char *filename);

   //! write strip to PLY file format
   void writePLYfile(const char *filename);

   //! convert strip from GEO to PLY file format
   static BOOLINT convertGEO2PLYfile(const char *filename,const char *outname);

   protected:

   unsigned int SIZE;
   unsigned int MAXSIZE;

   BOOLINT *ENDARRAY;

   float *VTXARRAY; // 3 components

   float *COLARRAY; // 3 or 4 components
   int COLCOMPS;

   float *NRMARRAY; // 3 components
   int NRMCOMPS;

   float *TEXARRAY; // 1-4 components
   int TEXCOMPS;

   float VTXX,VTXY,VTXZ;
   miniv3d BBOXMIN,BBOXMAX;

   float COLR,COLG,COLB,COLA;
   float COLAMIN,COLAMAX;

   float NRMX,NRMY,NRMZ;
   float TEXX,TEXY,TEXZ,TEXW;

   unsigned int COPYVTX;

   float SCALE;

   double MTX[16];
   BOOLINT MTXSET;
   double TEXMTX[16];
   BOOLINT TEXMTXSET;

   float ZSCALE;

   BOOLINT AUTODISABLECULLING;

   void addvtx();

   private:

   void init(int colcomps,int nrmcomps,int texcomps,unsigned int maxsize=1);
   void reinit(int colcomps,int nrmcomps,int texcomps,unsigned int maxsize=1);

   BOOLINT getnextrange(unsigned int &start,unsigned int &stop);

   int USESHADER;
   };

#endif
