// (c) by Stefan Roettger, licensed under LGPL 2.1

#include "minimtx.h"

#include "miniwls.h"

void miniWLS::linear_fit()
   {
   unsigned int n=size();

   vec2 lhs1(0,0);
   vec2 lhs2(0,0);
   vec2 rhs(0,0);

   // compute left- and right-hand side
   for (unsigned int i=0; i<n; i++)
      {
      WLS_Point p=at(i);

      double t=p.t;
      double t2=t*t;
      double m=p.m;
      double w2=p.w*p.w;

      lhs1.x+=w2;
      lhs1.y+=w2*t;

      lhs2.x+=w2*t;
      lhs2.y+=w2*t2;

      rhs.x+=w2*m;
      rhs.y+=w2*m*t;
      }

   mat2 M=mat2(lhs1,lhs2);
   double det=M.det();

   if (det!=0.0)
      c=M.invert()*rhs; // solve system of linear equations to fit linear polynomial
   else
      {
      double c0=rhs.x;
      double w2=lhs1.x;
      if (w2!=0.0) c0/=w2;
      c=vec3(c0,0); // fallback to constant polynomial
      }
   }

void miniWLS::quadratic_fit()
   {
   unsigned int n=size();

   vec3 lhs1(0,0,0);
   vec3 lhs2(0,0,0);
   vec3 lhs3(0,0,0);
   vec3 rhs(0,0,0);

   // accumulate left- and right-hand side
   for (unsigned int i=0; i<n; i++)
      {
      WLS_Point p=at(i);

      double t=p.t;
      double t2=t*t;
      double t3=t2*t;
      double t4=t3*t;
      double m=p.m;
      double w2=p.w*p.w;

      lhs1.x+=w2;
      lhs1.y+=w2*t;
      lhs1.z+=w2*t2;

      lhs2.x+=w2*t;
      lhs2.y+=w2*t2;
      lhs2.z+=w2*t3;

      lhs3.x+=w2*t2;
      lhs3.y+=w2*t3;
      lhs3.z+=w2*t4;

      rhs.x+=w2*m;
      rhs.y+=w2*m*t;
      rhs.z+=w2*m*t2;
      }

   mat3 M=mat3(lhs1,lhs2,lhs3);
   double det=M.det();

   if (det!=0.0)
      c=M.invert()*rhs; // solve system of linear equations to fit quadratic polynomial
   else
      {
      mat2 M=mat2(vec2(lhs1),vec2(lhs2));
      double det=M.det();

      if (det!=0.0)
         c=vec3(M.invert()*vec2(rhs)); // fallback to linear polynomial
      else
         {
         double c0=rhs.x;
         double w2=lhs1.x;
         if (w2!=0.0) c0/=w2;
         c=vec3(c0,0,0); // fallback to constant polynomial
         }
      }
   }

void miniWLS3D::linear_fit()
   {
   typedef minimtx<double,7> Matrix;

   unsigned int n=size();

   Matrix lhs(6,6),rhs(1,6),r;

   // accumulate left- and right-hand side
   for (unsigned int i=0; i<n; i++)
      {
      WLS_Point3D p=at(i);

      double t=p.t;
      double t2=t*t;
      vec3 m=p.m;
      double w2=p.w*p.w;

      lhs.add(0,0,w2);
      lhs.add(1,1,w2);
      lhs.add(2,2,w2);

      lhs.add(3,0,w2*t);
      lhs.add(4,1,w2*t);
      lhs.add(5,2,w2*t);

      lhs.add(0,3,w2*t);
      lhs.add(1,4,w2*t);
      lhs.add(2,5,w2*t);

      lhs.add(3,3,w2*t2);
      lhs.add(4,4,w2*t2);
      lhs.add(5,5,w2*t2);

      rhs.add(0,0,w2*m.x);
      rhs.add(0,1,w2*m.y);
      rhs.add(0,2,w2*m.z);
      rhs.add(0,3,w2*m.x*t);
      rhs.add(0,4,w2*m.y*t);
      rhs.add(0,5,w2*m.z*t);
      }

   if (lhs.solve(rhs,r)) // solve system of linear equations to fit linear parametric curve
      {
      a=vec3(r.get(0,0),r.get(0,1),r.get(0,2));
      b=vec3(r.get(0,3),r.get(0,4),r.get(0,5));
      c=vec3(0,0,0);
      }
   else
      a=b=c=vec3(0,0,0);

   if (b==vec3(0,0,0))
      {
      vec3 c(0,0,0);
      double wt=0.0;
      double w=0.0;

      for (unsigned int i=0; i<n; i++)
         {
         WLS_Point3D p=at(i);

         double t=p.t;
         vec3 m=p.m;
         double w2=p.w*p.w;

         c=c+w2*m;
         wt+=w2*t;
         w+=w2;
         }

      if (w>0.0)
         {
         c=c/w;
         wt/=w;
         }

      vec3 d(0,0,0);
      double wt2=0.0;

      for (unsigned int i=0; i<n; i++)
         {
         WLS_Point3D p=at(i);

         double t=p.t-wt;
         double t2=t*t;
         vec3 m=p.m-c;
         double w2=p.w*p.w;

         d=d+w2*m*t;
         wt2+=w2*t2;
         }

      b=d;
      if (wt2>0.0) b=b/wt2;
      a=c-wt*b;
      }
   }

void miniWLS3D::parabolic_fit()
   {
   typedef minimtx<double,10> Matrix;

   unsigned int n=size();

   Matrix lhs(9,9),rhs(1,9),r;

   // accumulate left- and right-hand side
   for (unsigned int i=0; i<n; i++)
      {
      WLS_Point3D p=at(i);

      double t=p.t;
      double t2=t*t;
      double t3=t2*t;
      double t4=t3*t;
      vec3 m=p.m;
      double w2=p.w*p.w;

      lhs.add(0,0,w2);
      lhs.add(1,1,w2);
      lhs.add(2,2,w2);

      lhs.add(3,0,w2*t);
      lhs.add(4,1,w2*t);
      lhs.add(5,2,w2*t);

      lhs.add(0,3,w2*t);
      lhs.add(1,4,w2*t);
      lhs.add(2,5,w2*t);

      lhs.add(3,3,w2*t2);
      lhs.add(4,4,w2*t2);
      lhs.add(5,5,w2*t2);

      lhs.add(6,0,w2*t2);
      lhs.add(7,1,w2*t2);
      lhs.add(8,2,w2*t2);

      lhs.add(0,6,w2*t2);
      lhs.add(1,7,w2*t2);
      lhs.add(2,8,w2*t2);

      lhs.add(3,6,w2*t3);
      lhs.add(4,7,w2*t3);
      lhs.add(5,8,w2*t3);

      lhs.add(6,3,w2*t3);
      lhs.add(7,4,w2*t3);
      lhs.add(8,5,w2*t3);

      lhs.add(6,6,w2*t4);
      lhs.add(7,7,w2*t4);
      lhs.add(8,8,w2*t4);

      rhs.add(0,0,w2*m.x);
      rhs.add(0,1,w2*m.y);
      rhs.add(0,2,w2*m.z);
      rhs.add(0,3,w2*m.x*t);
      rhs.add(0,4,w2*m.y*t);
      rhs.add(0,5,w2*m.z*t);
      rhs.add(0,6,w2*m.x*t2);
      rhs.add(0,7,w2*m.y*t2);
      rhs.add(0,8,w2*m.z*t2);
      }

   if (lhs.solve(rhs,r)) // solve system of linear equations to fit parabolic parametric curve
      {
      a=vec3(r.get(0,0),r.get(0,1),r.get(0,2));
      b=vec3(r.get(0,3),r.get(0,4),r.get(0,5));
      c=vec3(r.get(0,6),r.get(0,7),r.get(0,8));
      }
   else
      a=b=c=vec3(0,0,0);

   if (b==vec3(0,0,0) && c==vec3(0,0,0))
      linear_fit(); // linear fallback
   }
